import math
import random
from operator import attrgetter
from typing import Dict, List, Tuple

import numpy as np
import seaborn as sns
from deap import tools
from matplotlib import pyplot as plt
from pandas import DataFrame

from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome, convert_chromosome_to_schedule
from sampo.scheduler.genetic.operators import init_toolbox, ChromosomeType
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.schedule import ScheduleWorkDict, Schedule
from sampo.utilities.collections import reverse_dictionary


def build_schedule(wg: WorkGraph,
                   contractors: List[Contractor],
                   agents: WorkerContractorPool,
                   population_size: int,
                   generation_number: int,
                   selection_size: int,
                   mutate_order: float,
                   mutate_resources: float,
                   init_schedules: Dict[str, Schedule],
                   rand: random.Random,
                   spec: ScheduleSpec,
                   work_estimator: WorkTimeEstimator = None,
                   show_fitness_graph: bool = False) \
        -> ScheduleWorkDict:
    """
    Genetic algorithm
    Structure of chromosome:
    [[order of job], [numbers of workers types 1 for each job], [numbers of workers types 2], ... ]
    Different mate and mutation for order and for workers
    Generate order of job by prioritization from HEFT and from Topological
    Generate resources from min to max
    Overall initial population is valid
    :param show_fitness_graph:
    :param agents:
    :param work_estimator:
    :param contractors:
    :param wg:
    :param population_size:
    :param generation_number:
    :param selection_size:
    :param mutate_order:
    :param mutate_resources:
    :param rand:
    :param spec: spec for current scheduling
    :param init_schedules:
    :return: scheduler
    """

    if show_fitness_graph:
        fitness_history = list()

    index2node: Dict[int, GraphNode] = {index: node for index, node in enumerate(wg.nodes)}
    work_id2index: Dict[str, int] = {node.id: index for index, node in index2node.items()}
    worker_name2index = {worker_name: index for index, worker_name in enumerate(agents)}
    index2contractor = {ind: contractor.id for ind, contractor in enumerate(contractors)}
    index2contractor_obj = {ind: contractor for ind, contractor in enumerate(contractors)}

    init_chromosomes: Dict[str, ChromosomeType] = \
        {name: convert_schedule_to_chromosome(index2node, work_id2index, worker_name2index,
                                              reverse_dictionary(index2contractor), schedule)
         for name, schedule in init_schedules.items()}

    resources_border = np.zeros((2, len(agents), len(index2node)))
    for work_index, node in index2node.items():
        for req in node.work_unit.worker_reqs:
            worker_index = worker_name2index[req.kind]
            resources_border[0, worker_index, work_index] = req.min_count
            resources_border[1, worker_index, work_index] = \
                min(req.max_count, max(list(map(attrgetter('count'), agents[req.kind].values()))))

    toolbox = init_toolbox(wg, contractors, agents, index2node,
                           work_id2index, worker_name2index, index2contractor,
                           index2contractor_obj, init_chromosomes, mutate_order,
                           mutate_resources, selection_size, rand, spec, work_estimator)
    # save best individuals
    hof = tools.HallOfFame(1, similar=compare_individuals)
    # create population of a given size
    pop = toolbox.population(n=population_size)

    # probability to participate in mutation and crossover for each individual
    cxpb, mutpb = 0.5, 0.5
    mutpb_res, cxpb_res = 0.3, 0.3

    # map to each individual fitness function
    fitness = list(map(toolbox.evaluate, pop))
    for ind, fit in zip(pop, fitness):
        ind.fitness.values = [fit]

    if show_fitness_graph:
        fitness_history.append(sum(fitness) / len(fitness))

    g = 0

    while g < generation_number:
        # print("-- Generation %i --" % g)
        # select individuals of next generation
        offspring = toolbox.select(pop, len(pop))
        # clone selected individuals
        offspring = list(map(toolbox.clone, offspring))

        # operations for ORDER
        # crossover
        # take 2 individuals as input 1 modified individuals
        # take after 1: (1,3,5) and (2,4,6) and get pairs 1,2; 3,4; 5,6
        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            if rand.random() < cxpb:
                toolbox.mate(child1[0][0], child2[0][0])
                del child1.fitness.values  # remove previous
                del child2.fitness.values

        # mutation
        # take 1 individuals as input and return 1 individuals as output
        for mutant in offspring:
            if rand.random() < mutpb:
                toolbox.mutate(mutant[0][0])
                del mutant.fitness.values

        # gather all the fitness in one list and print the stats
        # invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        # evaluation for each individual
        # fitness = map(toolbox.evaluate, invalid_ind)
        # for ind, fit in zip(invalid_ind, fitness):
        #     ind.fitness.values = [fit]

        # renewing population
        pop[:] = offspring
        hof.update(pop)

        # operations for RESOURCES
        # mutation
        # select types for mutation
        # numbers of changing types
        number_of_type_for_changing = rand.randint(1, len(worker_name2index) - 1)
        # workers type for changing(+1 means contractor 'resource')
        workers = rand.sample(range(len(worker_name2index) + 1), number_of_type_for_changing)

        # resources mutation
        for worker in workers:
            low = resources_border[0, worker] if worker != len(worker_name2index) else 0
            up = resources_border[1, worker] if worker != len(worker_name2index) else 0
            for mutant in offspring:
                if rand.random() < mutpb_res:
                    toolbox.mutate_resources(mutant[0], low=low, up=up, type_of_worker=worker)
                    del mutant.fitness.values

        # for the crossover, we use those types that did not participate in the mutation(+1 means contractor 'resource')
        workers_for_mate = list(set(list(range(len(worker_name2index) + 1))) - set(workers))
        # crossover
        for ind_worker in range(len(workers_for_mate)):
            # take 2 individuals as input 1 modified individuals
            for child1, child2 in zip(offspring[::2], offspring[1::2]):
                if rand.random() < cxpb_res:
                    toolbox.mate_resources(child1[0][1][workers_for_mate[ind_worker]],
                                           child2[0][1][workers_for_mate[ind_worker]])
                    del child1.fitness.values  # remove prev
                    del child2.fitness.values

        # Gather all the fitness in one list and print the stats
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        # for each individual - evaluation
        fitness = map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitness):
            ind.fitness.values = [fit]

        if show_fitness_graph:
            _ftn = [f for f in fitness if not math.isinf(f)]
            if len(_ftn) > 0:
                fitness_history.append(sum(_ftn) / len(_ftn))

        # renewing population
        pop[:] = offspring
        hof.update(pop)

        # best = hof[0]
        # fits = [ind.fitness.values[0] for ind in pop]
        # evaluation = chromosome_evaluation(best, index2node, resources_border, work_id2index, worker_name2index,
        #                                   parent2inseparable_son, agents)
        # print("fits: ", fits)
        # print(evaluation)
        g += 1

    chromosome = hof[0][0]

    scheduled_works = convert_chromosome_to_schedule(chromosome, agents, index2node,
                                                     worker_name2index,
                                                     index2contractor_obj,
                                                     spec, work_estimator)

    if show_fitness_graph:
        sns.lineplot(
            data=DataFrame.from_records([(g * 4, v) for g, v in enumerate(fitness_history)],
                                        columns=["Поколение", "Функция качества"]),
            x="Поколение",
            y="Функция качества",
            palette='r')
        plt.show()

    return scheduled_works


def compare_individuals(a: Tuple[ChromosomeType], b: Tuple[ChromosomeType]):
    return a[0][0] == b[0][0] and (a[0][1] == b[0][1]).all()



from collections import deque
from typing import Optional, List, Union, Tuple, Dict

from sortedcontainers import SortedList

from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import GraphNode
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.types import ScheduleEvent, EventType, WorkerName, ContractorName
from sampo.utilities.collections import build_index

MomentumResourceTimeline = Dict[WorkerName, SortedList[ScheduleEvent]]
MomentumTimeline = Dict[ContractorName, MomentumResourceTimeline]


def prepare_worker(worker_pool: WorkerContractorPool, wreq: WorkerReq, contractor_id: str, count_getter) -> Worker:
    """
    :param worker_pool:
    :param wreq:
    :param contractor_id:
    :param count_getter: function `(req, worker) -> need_worker_count`
    :return:
    """
    worker = worker_pool[wreq.kind][contractor_id].copy()
    worker.count = count_getter(wreq, worker)
    return worker


def create_timeline(tasks: List[GraphNode], contractors: List[Contractor]) -> MomentumTimeline:
    """
    This should create empty Timeline from given list of tasks and contractor list
    :param tasks:
    :param contractors:
    :return:
    """

    # using  time, seq_id and event_type we can guarantee that
    # there may be only one possible order in cases:
    # (a) when events have the same time
    # (in this cases we need both time and seq_id to properly handle
    # available_workers processing logic)
    # (b) when events have the same time and their start and end matches
    # (service tasks for instance may have zero length)
    def event_cmp(x: Union[ScheduleEvent, Time, Tuple[Time, int, int]]) -> Tuple[Time, int, int]:
        if isinstance(x, ScheduleEvent):
            if x.event_type is EventType.Initial:
                return Time(-1), -1, x.event_type.priority

            return x.time, x.seq_id, x.event_type.priority

        if isinstance(x, Time):
            return x, len(tasks), 1

        if isinstance(x, tuple):
            return x

        raise ValueError(f"Incorrect type of value: {type(x)}")

    # to efficiently search for time slots for tasks to be scheduled
    # we need to keep track of starts and ends of previously scheduled tasks
    # and remember how many workers of a certain type is available at this particular moment
    timeline: MomentumTimeline = {
        c.id: {
            w_name: SortedList(
                iterable=(ScheduleEvent(-1, EventType.Initial, Time(0), None, ws.count),),
                key=event_cmp
            )
            for w_name, ws in c.workers.items()
        }
        for c in contractors
    }

    return timeline


def schedule(task_index: int,
             node: GraphNode,
             node2swork: Dict[GraphNode, ScheduledWork],
             inseparable_chain: List[GraphNode],
             timeline: MomentumTimeline,
             passed_agents: List[Worker],
             contractor: Contractor,
             work_estimator: Optional[WorkTimeEstimator] = None):
    """
    Schedule given `node` with `inseparable_chain`, selecting contractor from `contractors`
    using multi-contractor subsystem and `get_worker_count` for resource selection
    :param contractor:
    :param passed_agents:
    :param task_index:
    :param node:
    :param node2swork:
    :param inseparable_chain:
    :param timeline:
    :param work_estimator:
    :return:
    """

    start_time, max_parent_time, exec_times = find_min_start_time(timeline[contractor.id], node, node2swork,
                                                                  inseparable_chain, passed_agents, work_estimator)

    make_and_cache_schedule(task_index, node, node2swork, inseparable_chain, timeline, passed_agents,
                            contractor, start_time, exec_times)


def make_and_cache_schedule(task_index: int,
                            node: GraphNode,
                            node2swork: Dict[GraphNode, ScheduledWork],
                            inseparable_chain: List[GraphNode],
                            timeline: MomentumTimeline,
                            passed_agents: List[Worker],
                            contractor: Contractor,
                            start_time: Time,
                            exec_times: Dict[GraphNode, Tuple[Time, Time]]):
    # 6. create a schedule entry for the task

    nodes_start_times: Dict[GraphNode, Time] = {n: max((node2swork[pnode].finish_time
                                                        if pnode in node2swork else Time(0)
                                                        for pnode in n.parents),
                                                       default=Time(0))
                                                for n in inseparable_chain}

    curr_time = start_time
    for i, chain_node in enumerate(inseparable_chain):
        _, node_time = exec_times[chain_node]

        lag_req = nodes_start_times[chain_node] - start_time - node_time
        node_lag = lag_req if lag_req > 0 else 0

        start_work = curr_time + node_lag
        swork = ScheduledWork(
            work_unit=chain_node.work_unit,
            start_end_time=(start_work, start_work + node_time),
            workers=passed_agents,
            contractor=contractor
        )
        curr_time += node_time + node_lag
        node2swork[chain_node] = swork

    update_timeline(task_index, node, node2swork, inseparable_chain, timeline, passed_agents)


def update_timeline(task_index: int,
                    node: GraphNode,
                    node2swork: Dict[GraphNode, ScheduledWork],
                    inseparable_chain: List[GraphNode],
                    timeline: MomentumTimeline,
                    chosen_workers: List[Worker]):
    """
    Inserts `chosen_workers` into the timeline with it's `inseparable_chain`
    :param task_index:
    :param node:
    :param node2swork:
    :param inseparable_chain:
    :param timeline:
    :param chosen_workers:
    :return:
    """
    # 7. for each worker's specialization of the chosen contractor being used by the task
    # we update counts of available workers on previously scheduled events
    # that lie between start and end of the task
    # Also, add events of the start and the end to worker's specializations
    # of the chosen contractor.

    # experimental logics lightening. debugging showed its efficiency.

    swork = node2swork[node]  # masking the whole chain ScheduleEvent with the first node
    start = swork.start_time
    end = node2swork[inseparable_chain[-1]].finish_time
    for w in chosen_workers:
        state = timeline[w.contractor_id][w.name]
        start_idx = state.bisect_right(start)
        end_idx = state.bisect_right(end)
        available_workers_count = state[start_idx - 1].available_workers_count
        # updating all events in between the start and the end of our current task
        for event in state[start_idx: end_idx]:
            # assert event.available_workers_count >= w.count
            event.available_workers_count -= w.count

        # assert available_workers_count >= w.count

        if start_idx < end_idx:
            event: ScheduleEvent = state[end_idx - 1]
            # assert state[0].available_workers_count >= event.available_workers_count + w.count
            end_count = event.available_workers_count + w.count
        else:
            # assert state[0].available_workers_count >= available_workers_count
            end_count = available_workers_count

        state.add(ScheduleEvent(task_index, EventType.Start, start, swork, available_workers_count - w.count))
        state.add(ScheduleEvent(task_index, EventType.End, end, swork, end_count))


def schedule_with_time_spec(task_index: int,
                            node: GraphNode,
                            node2swork: Dict[GraphNode, ScheduledWork],
                            inseparable_chain: List[GraphNode],
                            timeline: MomentumTimeline,
                            workers: List[Worker],
                            contractor: Contractor,
                            assigned_time: Optional[Time],
                            work_estimator: Optional[WorkTimeEstimator] = None):
    if assigned_time:
        st, _, _ = find_min_start_time(timeline[contractor.id], node, node2swork,
                                       inseparable_chain, workers, work_estimator)
        exec_times = {n: (Time(0), assigned_time // len(inseparable_chain))
                      for n in inseparable_chain}

        make_and_cache_schedule(task_index, node, node2swork, inseparable_chain, timeline,
                                workers, contractor, st, exec_times)
    else:
        schedule(task_index, node, node2swork, inseparable_chain, timeline,
                 workers, contractor, work_estimator)


def find_min_start_time(resource_timeline: MomentumResourceTimeline,
                        node: GraphNode,
                        node2swork: Dict[GraphNode, ScheduledWork],
                        inseparable_chain: List[GraphNode],
                        passed_agents: List[Worker],
                        work_estimator: Optional[WorkTimeEstimator]) \
        -> Tuple[Time, Time, Dict[GraphNode, Tuple[Time, Time]]]:
    """
    Computes start time, max parent time, contractor and exec times for given node
    :param passed_agents: list of passed workers. Should be IN THE SAME ORDER AS THE CORRESPONDING WREQS
    :param resource_timeline:
    :param node:
    :param node2swork:
    :param inseparable_chain:
    :param work_estimator:
    :return:
    """
    # 1. identify earliest possible start time by max parent's end time

    max_parent_time: Time = max((node2swork[pnode].finish_time for pnode in node.parents), default=Time(0))
    nodes_start_times: Dict[GraphNode, Time] = {n: max((node2swork[pnode].finish_time
                                                        if pnode in node2swork else Time(0)
                                                        for pnode in n.parents),
                                                       default=Time(0))
                                                for n in inseparable_chain}

    # 2. calculating execution time of the task

    exec_time: Time = Time(0)
    exec_times: Dict[GraphNode, Tuple[Time, Time]] = {}  # node: (lag, exec_time)
    for chain_node in inseparable_chain:
        passed_agents_new = [agent.copy() for agent in passed_agents]

        node_exec_time: Time = Time(0) if len(chain_node.work_unit.worker_reqs) == 0 else \
            chain_node.work_unit.estimate_static(passed_agents_new, work_estimator)
        lag_req = nodes_start_times[chain_node] - max_parent_time - exec_time
        lag = lag_req if lag_req > 0 else 0

        exec_times[chain_node] = lag, node_exec_time
        exec_time += lag + node_exec_time

    if len(passed_agents) == 0:
        return max_parent_time, max_parent_time, exec_times

    return _find_min_start_time(resource_timeline, inseparable_chain, max_parent_time, exec_time, passed_agents), \
           max_parent_time, exec_times


def _find_min_start_time(resource_timeline: Dict[str, SortedList[ScheduleEvent]],
                         inseparable_chain: List[GraphNode],
                         parent_time: Time,
                         exec_time: Time,
                         passed_agents: List[Worker]) -> Time:
    # if it is a service unit, than it can be satisfied by any contractor at any moment
    # because no real workers is going to be used to execute the task
    # however, we still should respect dependencies of the service task
    # and should start it only after all the dependencies tasks are done
    if all((node.work_unit.is_service_unit for node in inseparable_chain)):
        return parent_time

    # checking if the contractor can satisfy requirements for the task at all
    # we return None in cases when the task cannot be executed
    # even if it is scheduled to the very end, e.g. after the end of all other tasks
    # already scheduled to this contractor

    for node in inseparable_chain:
        for i, wreq in enumerate(node.work_unit.worker_reqs):
            initial_event: ScheduleEvent = resource_timeline[wreq.kind][0]
            assert initial_event.event_type is EventType.Initial
            # if this contractor initially has fewer workers of this type, then needed...
            if initial_event.available_workers_count < passed_agents[i].count:
                return Time.inf()

    # here we look for the earliest time slot that can satisfy all the worker's specializations
    # we do it in that manner because each worker specialization can be treated separately
    # e.g. requested for different tasks
    # We check only the first node since all inseparable nodes have same worker_reqs despite the difference in exec time
    queue = deque(inseparable_chain[0].work_unit.worker_reqs)

    start = parent_time
    scheduled_wreqs: List[WorkerReq] = []

    type2count: Dict[str, int] = build_index(passed_agents, lambda w: w.name, lambda w: w.count)

    i = 0
    while len(queue) > 0:
        if i > 0 and i % 50 == 0:
            print(f"Warning! Probably cycle in looking for diff workers: {i} iteration")
        i += 1

        wreq = queue.popleft()
        state = resource_timeline[wreq.kind]
        # we look for the earliest time slot starting from 'start' time moment
        # if we have found a time slot for the previous task,
        # we should start to find for the earliest time slot of other task since this new time
        found_start = find_earliest_time_slot(state, start, exec_time, type2count[wreq.kind])

        assert found_start >= start

        if len(scheduled_wreqs) == 0 or start == found_start:
            # we schedule the first worker's specialization or the next spec has the same start time
            # as the all previous ones
            scheduled_wreqs.append(wreq)
            start = max(found_start, start)
        else:
            # The current worker specialization can be started only later than
            # the previously found start time.
            # In this case we need to add back all previously scheduled wreq-s into the queue
            # to be scheduled again with the new start time (e.g. found start).
            # This process should reach its termination at least at the very end of this contractor's schedule.
            queue.extend(scheduled_wreqs)
            scheduled_wreqs.clear()
            scheduled_wreqs.append(wreq)
            start = max(found_start, start)

    return start


def find_earliest_time_slot(
        state: SortedList[ScheduleEvent],
        parent_time: Time,
        exec_time: Time,
        required_worker_count: int) -> Time:
    current_start_time = parent_time
    current_start_idx = state.bisect_right(current_start_time) - 1

    # the condition means we have reached the end of schedule for this contractor subject to specialization (wreq)
    # as long as we assured that this contractor has enough capacity at all to handle the the task
    # we can stop and put the task at the very end
    i = 0
    while len(state[current_start_idx:]) > 0:
        if i > 0 and i % 50 == 0:
            print(f"Warning! Probably cycle in looking for earliest time slot: {i} iteration")
            print(f"Current start time: {current_start_time}, current start idx: {current_start_idx}")
        i += 1
        end_idx = state.bisect_right(current_start_time + exec_time)

        # checking from the end of execution interval, i.e., end_idx - 1
        # up to (including) the event right prepending the start of the execution interval, i.e., current_start_idx - 1
        # we need to check the event current_start_idx - 1 cause it is the first event
        # that influence amount of available for us workers
        not_enough_workers_found = False
        for idx in range(end_idx - 1, current_start_idx - 2, -1):
            if state[idx].available_workers_count < required_worker_count or state[idx].time < parent_time:
                # we're trying to find a new slot that would start with
                # either the last index passing the quantity check
                # or the index after the execution interval
                # we need max here to process a corner case when the problem arises
                # on current_start_idx - 1
                # without max it would get into infinite cycle
                current_start_idx = max(idx, current_start_idx) + 1
                not_enough_workers_found = True
                break

        if not not_enough_workers_found:
            break

        if current_start_idx >= len(state):
            break

        current_start_time = state[current_start_idx].time

    return current_start_time



import random
from typing import Dict, List, Tuple, Optional

from sampo.scheduler.resource.identity import IdentityResourceOptimizer
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.scheduler.base import Scheduler, SchedulerType
from sampo.scheduler.genetic.schedule_builder import build_schedule
from sampo.scheduler.heft.base import HEFTScheduler
from sampo.scheduler.heft_between.base import HEFTBetweenScheduler
from sampo.schemas.contractor import Contractor, get_worker_contractor_pool
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.utilities.validation import validate_schedule


class GeneticScheduler(Scheduler):

    def __init__(self,
                 number_of_generation: Optional[int] = 50,
                 size_selection: Optional[int or None] = None,
                 mutate_order: Optional[float or None] = None,
                 mutate_resources: Optional[float or None] = None,
                 size_of_population: Optional[float or None] = None,
                 rand: Optional[random.Random] = None,
                 seed: Optional[float or None] = None,
                 scheduler_type: SchedulerType = SchedulerType.Genetic,
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=IdentityResourceOptimizer(),
                         work_estimator=work_estimator)
        self.number_of_generation = number_of_generation
        self.size_selection = size_selection
        self.mutate_order = mutate_order
        self.mutate_resources = mutate_resources
        self.size_of_population = size_of_population
        self.rand = rand or random.Random(seed)
        self.work_estimator = work_estimator

    def get_params(self, works_count: int) -> Tuple[int, float, float, int]:
        size_selection = self.size_selection
        if size_selection is None:
            if works_count < 300:
                size_selection = 20
            else:
                size_selection = works_count // 15

        mutate_order = self.mutate_order
        if mutate_order is None:
            if works_count < 300:
                mutate_order = 0.006
            else:
                mutate_order = 2 / works_count

        mutate_resources = self.mutate_resources
        if mutate_resources is None:
            if works_count < 300:
                mutate_resources = 0.06
            else:
                mutate_resources = 18 / works_count

        size_of_population = self.size_of_population
        if size_of_population is None:
            if works_count < 300:
                size_of_population = 80
            elif 1500 > works_count >= 300:
                size_of_population = 50
            else:
                size_of_population = works_count // 50
        return size_selection, mutate_order, mutate_resources, size_of_population

    def schedule(self, wg: WorkGraph,
                 contractors: List[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 validate: bool = False) \
            -> Schedule:
        def init_schedule(scheduler_class):
            return scheduler_class(self.work_estimator).schedule(wg, contractors)

        init_schedules: Dict[str, Schedule] = {
            "heft_end": init_schedule(HEFTScheduler),
            "heft_between": init_schedule(HEFTBetweenScheduler)
        }

        size_selection, mutate_order, mutate_resources, size_of_population = self.get_params(wg.vertex_count)
        agents = get_worker_contractor_pool(contractors)

        scheduled_works = build_schedule(wg,
                                         contractors,
                                         agents,
                                         size_of_population,
                                         self.number_of_generation,
                                         size_selection,
                                         mutate_order,
                                         mutate_resources,
                                         init_schedules,
                                         self.rand,
                                         spec,
                                         self.work_estimator)
        schedule = Schedule.from_scheduled_works(scheduled_works.values(), wg)

        if validate:
            validate_schedule(schedule, wg, contractors)

        return schedule



from collections import defaultdict
from dataclasses import dataclass
from typing import List, Dict, Union
from uuid import uuid4

import numpy as np
from pandas import DataFrame, Series

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.resources import Worker, Equipment
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.types import WorkerName, ContractorName
from sampo.utilities.serializers import custom_serializer

WorkerContractorPool = Dict[WorkerName, Dict[ContractorName, Worker]]
DefaultContractorCapacity = 25


@dataclass
class Contractor(AutoJSONSerializable['Contractor'], Identifiable):
    """
    Used to store information about the contractor and its resources
    :param workers: dictionary, where the key is the employee's specialty, and the value is the pool of employees of
    this specialty
    :param equipments: dictionary, where the key is the type of technique, and the value is the pool of techniques of
    that type
    """
    workers: Dict[str, Worker]
    equipments: Dict[str, Equipment]

    def __post_init__(self):
        for w in self.workers.values():
            w.contractor_id = self.id

    @custom_serializer('workers')
    def serialize_workers(self, value):
        return [{'key': k, 'val': v._serialize()} for k, v in value.items()]

    @custom_serializer('equipments')
    def serialize_equipment(self, value):
        return {k: v._serialize() for k, v in value.items()}

    @classmethod
    @custom_serializer('workers', deserializer=True)
    def deserialize_workers(cls, value):
        return {tuple(i['key']): Worker._deserialize(i['val']) for i in value}

    @classmethod
    @custom_serializer('equipments', deserializer=True)
    def deserialize_equipment(cls, value):
        return {k: Equipment._deserialize(v) for k, v in value.items()}


def get_worker_contractor_pool(contractors: Union[List['Contractor'], 'Contractor']) -> WorkerContractorPool:
    """
    Gets agent dictionary from contractors list.
    Alias for frequently used functionality.
    :param contractors: List of all the considered contractors
    :return: Dictionary of workers by worker name, next by contractor id
    """
    agents = defaultdict(dict)
    for contractor in contractors:
        for name, worker in contractor.workers.items():
            agents[name][contractor.id] = worker.copy()
    return agents


def get_contractor_for_resources_schedule(resources: Union[DataFrame, List[Dict[str, int]]],
                                          contractor_capacity: int = DefaultContractorCapacity,
                                          contractor_id: str = None,
                                          contractor_name: str = "") \
        -> 'Contractor':
    """
    Generates a contractor, which satisfies the provided resources
    :param resources: List of resource requirements for each task
    :param contractor_id: ID of the new contractor. If None, a new UUID4 is generated
    :param contractor_name: Name of the new contractor. If None, an empty string should be passed
    :param contractor_capacity: Capacity of the generated contractor. It influences scale of resource capacities
    :return: A new contractor of the given type
    """

    def aggregate_resources(resources_in: DataFrame, contractor_capacity_in: int) -> Series:
        """
        Aggregates resources based on the contractor's capacity
        :param resources_in: DataFrame with the (possibly sparse) resource grid
        :param contractor_capacity_in: Capacity of the generated contractor
        :return: Series with resources count
        """
        min_contractor = resources_in.max(axis=0, skipna=True, numeric_only=True)

        return np.ceil(min_contractor * contractor_capacity_in / 10)

    resources_df = resources if isinstance(resources, DataFrame) else DataFrame(resources)
    contractor_id = contractor_id if contractor_id else str(uuid4().hex)

    # TODO: process equipment
    return Contractor(id=contractor_id,
                      name=contractor_name,
                      workers={item[0]: Worker(str(i), item[0], int(item[1]), contractor_id)
                               for i, item in enumerate(aggregate_resources(resources_df, contractor_capacity).items())},
                      equipments=dict())



from typing import List, Dict, Set, Optional, Iterable

import numpy as np
from toposort import toposort_flatten, toposort

from sampo.scheduler.base import Scheduler
from sampo.scheduler.base import SchedulerType
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.utils.momentum_timeline import create_timeline, schedule_with_time_spec
from sampo.scheduler.utils.multi_contractor import get_best_contractor_and_worker_borders
from sampo.schemas.contractor import Contractor, get_worker_contractor_pool
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.validation import validate_schedule


class TopologicalScheduler(Scheduler):

    def __init__(self, scheduler_type: SchedulerType = SchedulerType.Topological,
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=AverageReqResourceOptimizer(),
                         work_estimator=work_estimator)

    def schedule(self, wg: WorkGraph,
                 contractors: List[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 validate: bool = False) \
            -> Schedule:
        # Checking pre-conditions for this scheduler_topological to be applied
        # check_all_workers_have_same_qualification(wg, contractors)

        tsorted_nodes: List[GraphNode] = self._topological_sort(wg)

        schedule = Schedule.from_scheduled_works(
            self.build_scheduler(tsorted_nodes, contractors, spec, self.work_estimator), wg
        )

        # check the validity received scheduler
        if validate:
            validate_schedule(schedule, wg, contractors)

        return schedule

    # noinspection PyMethodMayBeStatic
    def _topological_sort(self, wg: WorkGraph) -> List[GraphNode]:
        node2ind = {node: i for i, node in enumerate(wg.nodes)}

        dependents: Dict[int, Set[int]] = {
            node2ind[v]: {node2ind[node] for node in v.parents}
            for v in wg.nodes
        }

        tsorted_nodes_indices: List[int] = toposort_flatten(dependents, sort=True)
        tsorted_nodes = [wg.nodes[i] for i in tsorted_nodes_indices]
        return tsorted_nodes

    # noinspection PyMethodMayBeStatic
    def build_scheduler(self, tasks: List[GraphNode], contractors: List[Contractor],
                        spec: ScheduleSpec,
                        work_estimator: WorkTimeEstimator = None) \
            -> Iterable[ScheduledWork]:
        """
        Builds a schedule from a list of tasks where all dependents tasks are guaranteed to be later
        in the sequence than their dependencies
        :param work_estimator:
        :param tasks: list of tasks ordered by some algorithm according to their dependencies and priorities
        :param spec: spec for current scheduling
        :param contractors: pools of workers available for execution
        :return: a schedule
        """

        # data structure to hold scheduled tasks
        node2swork: Dict[GraphNode, ScheduledWork] = dict()

        timeline = create_timeline(tasks, contractors)

        # we can get agents here, because they are always same and not updated
        worker_pool = get_worker_contractor_pool(contractors)

        # We allocate resources for the whole inseparable chain, when we process the first node in it.
        # So, we will store IDs of non-head nodes in such chains to skip them.
        # Note that tasks are already topologically ordered,
        # i.e., the first node in a chain is always processed before its children

        skipped_inseparable_children: Set[str] = set()
        # scheduling all the tasks in a one-by-one manner
        for index, node in enumerate(tasks):
            # skip, if this node was processed as a part of an inseparable chin previously
            if node.id in skipped_inseparable_children:
                continue
            work_unit = node.work_unit
            work_spec = spec.get_work_spec(work_unit.id)

            # 0. find, if the node starts an inseparable chain

            inseparable_chain = node.get_inseparable_chain()
            if inseparable_chain:
                skipped_inseparable_children.update((ch.id for ch in inseparable_chain))
            whole_work_nodes = inseparable_chain if inseparable_chain else [node]

            min_count_worker_team, max_count_worker_team, contractor, workers = \
                get_best_contractor_and_worker_borders(worker_pool, contractors, node.work_unit.worker_reqs)

            best_worker_team = [worker.copy() for worker in workers]

            # apply worker team spec
            self.optimize_resources_using_spec(work_unit, best_worker_team, work_spec,
                                               lambda optimize_array: self.resource_optimizer.optimize_resources(
                                                   worker_pool, best_worker_team,
                                                   optimize_array,
                                                   min_count_worker_team, max_count_worker_team,
                                                   # dummy
                                                   lambda _: Time(0)))

            # finish scheduling with time spec
            schedule_with_time_spec(index, node, node2swork, whole_work_nodes, timeline, best_worker_team, contractor,
                                    work_spec.assigned_time, work_estimator)

        return node2swork.values()


class RandomizedTopologicalScheduler(TopologicalScheduler):
    def __init__(self, work_estimator: Optional[WorkTimeEstimator or None] = None,
                 random_seed: Optional[int] = None):
        super().__init__(work_estimator)
        self._random_state = np.random.RandomState(random_seed)

    def _topological_sort(self, wg: WorkGraph) -> List[GraphNode]:
        def shuffle(nodes: Set[GraphNode]) -> List[GraphNode]:
            nds = list(nodes)
            indices = np.arange(len(nds))
            self._random_state.shuffle(indices)
            return [nds[ind] for ind in indices]

        dependents: Dict[GraphNode, Set[GraphNode]] = {v: v.parents_set for v in wg.nodes}
        tsorted_nodes: List[GraphNode] = [
            node for level in toposort(dependents)
            for node in shuffle(level)
        ]

        # print([node.work_unit.id for node in tsorted_nodes])
        return tsorted_nodes



import pandas as pd

from sampo.schemas.time import Time


def fix_split_tasks(baps_schedule_df: pd.DataFrame) -> pd.DataFrame:
    """
    Process and merge information for all tasks, which were separated on the several stages during split
    :param baps_schedule_df: pd.DataFrame: schedule with info for tasks separated on stages
    :return: pd.DataFrame: schedule with merged info for all real tasks
    """
    df_len = baps_schedule_df.shape[0]
    baps_schedule_df.index = range(df_len)

    df = pd.DataFrame(columns=baps_schedule_df.columns)
    unique_ids = set([x.split('_')[0] for x in baps_schedule_df.task_id])

    for task_id in unique_ids:
        task_stages_df = baps_schedule_df.loc[baps_schedule_df.loc[:, 'task_id'].str.contains(task_id)]
        task_series = merge_split_stages(task_stages_df.reset_index(drop=True))
        df.loc[df.shape[0]] = task_series  # append

    df = df.sort_values(by=['start', 'task_name'])
    df['idx'] = range(len(df))

    return df


def merge_split_stages(task_df: pd.DataFrame) -> pd.Series:
    """
    Merge split stages of the same real task into one
    :param task_df: pd.DataFrame: one real task's stages dataframe, sorted by start time
    :return: pd.Series with the full information about the task
    """
    if len(task_df) == 1:
        return task_df.loc[0, :]
    else:
        df = task_df.iloc[-1:].reset_index(drop=True)
        for column in ['task_id', 'task_name']:
            df.loc[0, column] = df.loc[0, column].split('_')[0]  # fix task id and name
        # fix task's start time and duration
        df.loc[0, 'start'] = task_df.loc[0, 'start']
        df.loc[0, 'finish'] = task_df.loc[len(task_df) - 1, 'finish']
        df.loc[0, 'duration'] = (df.loc[0, 'finish'] - df.loc[0, 'start']).days + 1
        return df.loc[0, :]
